// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include IO utilities:
#include <iostream>

// Include Vec4f class:
#include "Vec4f.h"

// Include Mat4f class:
#include "Mat4f.h"

// Include our shader wrapper class:
#include "Shader.h"

// Include our mesh class:
#include "Mesh.h"


/**
 * Initializes the application including the GUI which you can
 * use to test the Vec4f and Mat4f functionaliy.
 */

int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(1090, 560, "Hello Triangle Framework (Computergrafik 1)", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf" , 16.0f);

    // Create a triangle:
    std::vector<Triangle> triangles;
    {
        // Create vertex a of triangle:
        Vertex vertexA(Vec4f( 0.866f, -0.5f, 0.0f) * 0.5f,Vec4f(0,0,1,0),Vec4f(1,0,0));

        // Create vertex b of triangle:
        Vertex vertexB(Vec4f(0,1.f,0) * 0.5f,Vec4f(0,0,1,0),Vec4f(0,0,1));

        // Create vertex c of triangle:
        Vertex vertexC(Vec4f(-0.866f, -0.5f,0) * 0.5f,Vec4f(0,0,1,0),Vec4f(0,1,0));

        // Create triangle:
        Triangle triangle(vertexA, vertexB, vertexC);

        // Add triangle to the list of triangles:
        triangles.push_back(triangle);
    }

    // Create our mesh containing one triangle:
    Mesh triangleMesh(triangles);

    // Load our simple shader:
    Shader shader(
                CMAKE_SOURCE_DIR "/shader/simple.vert",
                CMAKE_SOURCE_DIR "/shader/simple.frag"
    );

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Start the Dear ImGui frame
        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Setup the GL viewport
        glViewport(0, 0, display_w, display_h);
        glClearColor(0.6f, 0.725f, 0.8f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT);

        // Bind our shader:
        shader.bind();

        // Draw our triangle:
        triangleMesh.render();

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

